from typing import Optional, List
from thinc.types import Floats2d
from thinc.api import Model

from spacy.util import registry
from spacy.tokens import Doc


def EntityLinker_v1(
    tok2vec: Model, nO: Optional[int] = None
) -> Model[List[Doc], Floats2d]:
    chain = registry.get("layers", "chain.v1")
    clone = registry.get("layers", "clone.v1")
    with Model.define_operators({">>": chain, "**": clone}):
        token_width = tok2vec.maybe_get_dim("nO")

        Linear = registry.get("layers", "Linear.v1")
        output_layer = Linear(nO=nO, nI=token_width)

        list2ragged = registry.get("layers", "list2ragged.v1")
        reduce_mean = registry.get("layers", "reduce_mean.v1")
        residual = registry.get("layers", "residual.v1")
        Maxout = registry.get("layers", "Maxout.v1")

        model = (
            tok2vec
            >> list2ragged()
            >> reduce_mean()
            >> residual(Maxout(nO=token_width, nI=token_width, nP=2, dropout=0.0))  # type: ignore[arg-type]
            >> output_layer
        )
        model.set_ref("output_layer", output_layer)
        model.set_ref("tok2vec", tok2vec)
    return model



import pytest
from spacy.training import Example
from spacy.lang.en import English
from thinc.api import Config


default_tok2vec_config = """
[model]
@architectures = "spacy-legacy.HashEmbedCNN.v1"
pretrained_vectors = null
width = 96
depth = 4
embed_size = 2000
window_size = 1
maxout_pieces = 3
subword_features = true
"""
DEFAULT_TOK2VEC_MODEL = Config().from_str(default_tok2vec_config)["model"]


TAGS = ("N", "V", "J")

TRAIN_DATA = [
    ("I like green eggs", {"tags": ["N", "V", "J", "N"]}),
    ("Eat blue ham", {"tags": ["V", "J", "N"]}),
]


@pytest.mark.parametrize(
    "tagger_config",
    [
        {
            "@architectures": "spacy-legacy.Tagger.v1",
            "tok2vec": DEFAULT_TOK2VEC_MODEL,
        }
    ],
)
def test_overfitting_IO(tagger_config):
    # Simple test to try and quickly overfit the tagger - ensuring the ML models work correctly
    pipe_config = {"model": tagger_config}
    nlp = English()
    tagger = nlp.add_pipe("tagger", config=pipe_config)
    train_examples = []
    for t in TRAIN_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))
    optimizer = nlp.initialize(get_examples=lambda: train_examples)
    assert tagger.model.get_dim("nO") == len(TAGS)

    for i in range(50):
        losses = {}
        nlp.update(train_examples, sgd=optimizer, losses=losses)
    assert losses["tagger"] < 0.00001

    # test the trained model
    test_text = "I like blue eggs"
    doc = nlp(test_text)
    assert doc[0].tag_ == "N"
    assert doc[1].tag_ == "V"
    assert doc[2].tag_ == "J"
    assert doc[3].tag_ == "N"
